from bisect import bisect
from collections import Hashable, UserDict
from numbers import Number
from typing import Dict, Tuple

CredibilityInterval = Tuple[float, float]


class CumulativeDistribution(UserDict):
    def likelihood(self, outcome: Hashable) -> float:
        (first_outcome, _), *outcomes = self.items()

        if outcome < first_outcome:
            return 0

        index = bisect(self.keys(), outcome) - 1
        return list(self.values())[index]

    def outcome(self, probability: float):
        assert (
            0 < probability > 1
        ), "Probability should be a number from [0, 1] interval"

        if probability == 0.0:
            return list(self.keys())[0]

        if probability == 1.0:
            return list(self.keys())[-1]

        index = bisect(self.values(), probability)

        if probability == list(self.values())[index - 1]:
            return list(self.keys())[index - 1]

        return list(self.keys())[index]

    def credible_interval(self, credibility: float) -> CredibilityInterval:
        tail_probability: float = (1 - credibility) / 2

        return self.outcome(tail_probability), self.outcome(1 - tail_probability)

    @classmethod
    def from_dict(cls, outcomes: Dict[Hashable, Number]):
        cumulative_outcomes = []
        cumulative_probabilities = []
        running_probability_sum: float = 0.0

        for outcome, probability in sorted(outcomes.items()):
            running_probability_sum += probability

            cumulative_outcomes.append(outcome)
            cumulative_probabilities.append(running_probability_sum)

        total_probability = running_probability_sum

        probabilities = [
            cumulative_probability / total_probability
            for cumulative_probability in cumulative_probabilities
        ]

        return cls(
            {
                outcome: probability
                for outcome, probability in zip(cumulative_outcomes, probabilities)
            }
        )



import random
from collections import UserDict
from numbers import Number
from typing import Callable, Hashable, Iterable, List, Mapping, Optional, Union, cast

from tabulate import tabulate

from posterior.distributions import CumulativeDistribution

Outcome = Union[str, float, int]
Probability = float

Comparable = Union[float, int, "RandomVariable"]
ProbabilityFunc = Callable[[Outcome], Probability]
InitialOutcomes = Union[Iterable[Outcome], Mapping[Outcome, Probability]]


class RandomVariable(UserDict[Outcome, Probability]):
    """
    Set of outcomes and corresponding probabilities/frequencies of a random variable
    """

    def __init__(
        self,
        outcomes: Optional[InitialOutcomes] = None,
        probability_func: ProbabilityFunc = lambda outcome: 1,
    ):
        if isinstance(outcomes, Iterable):
            outcomes = {outcome: probability_func(outcome) for outcome in outcomes}

        super(RandomVariable, self).__init__(outcomes)
        self.normalize()

    @property
    def total(self):
        return sum(self.values())

    @property
    def mean(self) -> float:
        if not all([isinstance(outcome, Number) for outcome in self.keys()]):
            raise ValueError("Outcomes are not numeric values. Cannot compute the mean")

        return sum(
            [
                cast(float, outcome) * probability
                for outcome, probability in self.items()
            ]
        )

    @property
    def variance(self) -> float:
        if not isinstance(list(self.keys()), Number):
            raise ValueError(
                "Outcomes are not numeric values. Cannot compute the variance"
            )

        mean = self.mean

        return sum(
            [
                probability * (cast(float, outcome) - mean) ** 2
                for outcome, probability in self.items()
            ]
        )

    @property
    def max_likelihood(self) -> Probability:
        return max(self.values())

    def random(self) -> List[Hashable]:
        return random.choices(population=list(self.keys()), weights=list(self.values()))

    def normalize(self, target_total: float = 1.0):
        total = self.total

        if not total:
            return

        norm_factor: float = target_total / total

        for outcome, probability in self.data.items():
            self.data[outcome] = probability * norm_factor

    def __lt__(self, item: Comparable):
        if isinstance(item, RandomVariable):
            # probability that self is less then item
            return sum(
                probability1 * probability2
                for outcome1, probability1 in self.items()
                for outcome2, probability2 in item.items()
                if outcome1 < outcome2
            )

        return sum(
            [probability for outcome, probability in self.items() if probability < item]
        )

    def __gt__(self, item: Comparable):
        if isinstance(item, RandomVariable):
            return sum(
                probability1 * probability2
                for outcome1, probability1 in self.items()
                for outcome2, probability2 in item.items()
                if outcome1 > outcome2
            )

        return sum(
            [probability for outcome, probability in self.items() if probability > item]
        )

    def cdf(self):
        return CumulativeDistribution(self)

    def __repr__(self) -> str:
        """Render component map as a table"""
        table = tabulate(
            self.items(),
            headers=("Outcomes", "Probabilities"),
            tablefmt="fancy_grid",
        )

        return f"{self.__class__.__name__}:\n{table}"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['posterior']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.3,<2.0.0', 'tabulate>=0.8.9,<0.9.0']

setup_kwargs = {
    'name': 'posterior',
    'version': '0.1.0',
    'description': "A lightweight framework for Bayes's inference",
    'long_description': None,
    'author': 'Roman Glushko',
    'author_email': 'roman.glushko.m@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from collections import Callable, MutableMapping
from numbers import Number
from typing import Any, Dict, Iterable, List, Tuple, Union

from posterior.random_variables import Outcome, Probability, RandomVariable

Hypothesis = Union[str, float, int]

Likelihood = Union[
    Dict[Hypothesis, Dict[Outcome, Probability]],
    Dict[Hypothesis, Any],
    Callable[[Hypothesis, Outcome], Probability],
]


def flatten_dict(
    dictionary: MutableMapping,
    parent_key: Union[bool, str] = False,
    separator: str = ".",
) -> dict:
    """
    Turns a nested dictionary into a flattened dictionary
    """

    items: List[Tuple[str, Any]] = []

    for key, value in dictionary.items():
        new_key = str(parent_key) + separator + key if parent_key else key

        if isinstance(value, MutableMapping):
            items.extend(flatten_dict(value, new_key, separator).items())
        elif isinstance(value, list):
            for k, v in enumerate(value):
                items.extend(flatten_dict({str(k): v}, new_key).items())
        else:
            items.append((new_key, value))

    return dict(items)


class Hypotheses(RandomVariable[Hypothesis, Probability]):
    def __init__(self, *args, likelihood: Likelihood, **kwargs):
        super().__init__(*args, **kwargs)

        if callable(likelihood):
            self.likelihood = likelihood
            return

        # if likelihood is dict
        self.likelihood = {
            hypo: flatten_dict(outcomes) for hypo, outcomes in likelihood.items()
        }

    def evaluate(self, new_data: Union[Outcome, Iterable[Outcome]]):
        """
        Considers evidence of new data to update believes about hypothesis
        """
        if isinstance(new_data, (Number, str)):
            new_data = [new_data]

        for data in new_data:
            for hypothesis, probability in self.items():
                self[hypothesis] = probability * self.get_likelihood(hypothesis, data)

        self.normalize()

    def get_likelihood(
        self, hypothesis: Hypothesis, new_outcome: Outcome
    ) -> Probability:
        """
        Computes likelihood of a hypothesis given evidence of the data
        """
        if callable(self.likelihood):
            return self.likelihood(hypothesis, new_outcome)

        return self._get_likelihood_from_dict(hypothesis, new_outcome)

    def _get_likelihood_from_dict(
        self, hypothesis: Hypothesis, new_outcome: Outcome
    ) -> Probability:
        hypothesis = self.likelihood.get(hypothesis, None)

        if not hypothesis:
            raise ValueError()

        probability = hypothesis.get(new_outcome, None)

        if probability is None:
            raise ValueError()

        return probability
